import threading
import pyperclip
from colorama import Fore


class CommandHandler:
    """
    Handles user input for VLCYT
    """

    _help_commands = ["?", "help"]
    _volume_commands = ["volume", "v"]
    _skip_commands = ["skip", "s", "next", "n", "forward", "f"]
    _play_commands = ["play", "pause", "p"]
    _repeat_commands = ["repeat", "replay", "r"]
    _back_commands = ["back", "b"]
    _loop_commands = ["loop", "l"]
    _shuffle_commands = ["shuffle"]
    _copy_url_commands = ["copy", "c", "url"]
    _exit_commands = ["exit", "quit", "q"]

    def __init__(self, vlcyt):
        self.vlcyt = vlcyt
        self.input_thread = threading.Thread(target=self._get_input)
        self.input_thread.daemon = True

    def _get_input(self):
        """
        Gathers user input from the input thread and executes commands.
        """
        while True:
            command_name, command_value = self._get_command()
            if command_name in self._help_commands:
                self.command_help()
            elif command_name in self._volume_commands:
                self.command_set_volume(command_value)
            elif command_name in self._skip_commands:
                self.command_skip_song(command_value)
            elif command_name in self._play_commands:
                self.command_p()
            elif command_name in self._repeat_commands:
                self.command_repeat()
            elif command_name in self._back_commands:
                self.command_back()
            elif command_name in self._loop_commands:
                self.command_loop()
            elif command_name in self._shuffle_commands:
                self.command_shuffle()
            elif command_name in self._copy_url_commands:
                self.command_copy_url()
            elif command_name in self._exit_commands:
                self.vlcyt.exit_program = True
            else:
                print(f"{Fore.RED}Invalid command{Fore.RESET}")

    def _get_command(self):
        """
        Processes a command from the user.
        Output: tuple: command name string, command value string
        """
        command = ""
        while command == "":
            command = input(self.vlcyt.command_string).lower()
        split_command = command.split()
        try:
            command_name = split_command[0]
            command_value = split_command[1]
        except IndexError:
            command_value = None
        return command_name, command_value

    def command_help(self):
        print(
            f"""
{Fore.MAGENTA}======================================
{Fore.YELLOW}NOTE: Most commands have multiple aliases separated by commas, use whichever you prefer.

{Fore.CYAN}VLCYT Commands:

{Fore.GREEN}?, help{Fore.WHITE}
Opens this help menu and shows whether or not looping and shuffling are enabled.

{Fore.GREEN}volume, v{Fore.WHITE}
Adjust the volume (0 - 100).

{Fore.GREEN}skip, s, next, n, forward, f{Fore.WHITE}
Skips song(s).
For example: Entering "skip" will skip one song,
entering "skip 5" will skip 5 songs.

{Fore.GREEN}play, pause, p{Fore.WHITE}
Plays/Pauses the current song.

{Fore.GREEN}repeat, replay, r{Fore.WHITE}
Repeats the current song one time.

{Fore.GREEN}back, b{Fore.WHITE}
Skips to the last song played.

{Fore.GREEN}loop, l{Fore.WHITE}
Enables looping.
The current song will keep playing until looping is disabled.

{Fore.GREEN}shuffle{Fore.WHITE}
Shuffles the playlist without repeating until every song has been played.

{Fore.GREEN}copy, c, url{Fore.WHITE}
Copy the current song's YouTube URL.

{Fore.GREEN}exit, quit, q{Fore.WHITE}
Closes the program.
{Fore.MAGENTA}======================================
{Fore.CYAN}---Settings---{Fore.RESET}
{Fore.GREEN}Looping:{Fore.RESET} {f"{Fore.GREEN}Enabled{Fore.RESET}" if self.vlcyt.loop_song else f"{Fore.RED}Disabled{Fore.RESET}"}
{Fore.GREEN}Shuffling:{Fore.RESET} {f"{Fore.GREEN}Enabled{Fore.RESET}" if self.vlcyt.shuffle_playlist else f"{Fore.RED}Disabled{Fore.RESET}"}
{Fore.CYAN}--------------{Fore.RESET}"""
        )

    def command_set_volume(self, volume):
        """
        Sets VLC volume.
        """
        try:
            volume = int(volume)
        except (TypeError, ValueError):
            print(f"{Fore.RED}Bad input.{Fore.RESET} Enter an integer from 0 - 100.")
            return
        if 0 <= volume <= 100:
            if (
                self.vlcyt.vlc_player.audio_set_volume(volume) == 0
            ):  # Returns 0 if volume was set
                print(f"Volume set to {Fore.GREEN}{volume}{Fore.RESET}")
        else:
            print(f"{Fore.RED}Volume out of range.{Fore.RESET} Range: 0 - 100")

    def command_skip_song(self, amount_to_skip):
        """
        Skips the current song.
        Round robins if it causes the song counter to go over the total amount of songs in the playlist.
        """

        if amount_to_skip is not None:
            try:
                amount_to_skip = int(amount_to_skip)
            except (TypeError, ValueError):
                print(f"{Fore.RED}Bad input.{Fore.RESET} Enter a number.")
                return

        if amount_to_skip in [1, None]:
            self.vlcyt.song_index = (
                self.vlcyt.song_index + 1
                if self.vlcyt.song_index < self.vlcyt.total_songs
                else 0
            )
            self.vlcyt.skip_song = True
        elif amount_to_skip > 1:
            potential_index = self.vlcyt.song_index + amount_to_skip
            if potential_index <= self.vlcyt.total_songs:
                self.vlcyt.song_index += amount_to_skip - 1
                self.vlcyt.skip_song = True
            else:  # Round robin
                total_multiplier = potential_index // self.vlcyt.total_songs
                self.vlcyt.song_index = (
                    potential_index - 1 - self.vlcyt.total_songs * total_multiplier
                )
                self.vlcyt.skip_song = True
        else:
            print(f"{Fore.RED}Bad input.{Fore.RESET} Enter a value greater than 0.")

    def command_repeat(self):
        """
        Repeats the current song.
        """
        self.vlcyt.vlc_player.set_time(0)

    def command_p(self):
        """
        Plays/Pauses the current song.
        """
        self.vlcyt.vlc_player.pause()

    def command_back(self):
        """
        Play last song in history.
        """
        if self.vlcyt.song_history:
            self.vlcyt.back_song = True
            self.vlcyt.skip_song = True
        else:
            print(f"{Fore.RED}No songs in history{Fore.RESET}")

    def command_loop(self):
        """
        Enables/Disables looping the current song.
        """
        if self.vlcyt.loop_song == False:
            self.vlcyt.loop_song = True
            print(f"Looping {Fore.GREEN}enabled.{Fore.RESET}")
        else:
            self.vlcyt.loop_song = False
            print(f"Looping {Fore.RED}disabled.{Fore.RESET}")

    def command_shuffle(self):
        if self.vlcyt.shuffle_playlist == False:
            self.vlcyt.shuffle_playlist = True
            print(f"Shuffle {Fore.GREEN}enabled.{Fore.RESET}")
        else:
            self.vlcyt.shuffle_playlist = False
            print(f"Shuffle {Fore.RED}disabled.{Fore.RESET}")
    
    def command_copy_url(self):
        pyperclip.copy("https://www.youtube.com/watch?v=" + self.vlcyt.current_song.videoid)
        print(f"{Fore.GREEN}Song URL Copied")



import argparse
import pafy
import os
import time
import sys
import threading
import random
from vlcyt.command_handler import CommandHandler
from colorama import Fore, Back, Style


class VLCYT:
    """
    Stream YouTube playlist audio in VLC behind the scenes from the command line.
    """

    command_string = f"{Fore.RESET}{Back.RESET}>"

    def __init__(self, playlist_url, song_info_enabled=True):
        self.playlist = pafy.get_playlist2(playlist_url)  # Pafy playlist object
        self.song_index = 0  # Current song index
        self.song_counter = 0  # Stores how many songs have been played, resets if every song has been played.
        self.total_songs = len(self.playlist)  # Amount of songs in the Pafy object
        self.current_song = None  # Stores the current song
        self.vlc_player = vlc.MediaPlayer()  # Stores the VLC object
        self.song_info_enabled = song_info_enabled  # The current song information is printed when the song changes if this is enabled

        self.song_history = []  # Stores indexes of songs that have been played

        # User input
        self.cmds = CommandHandler(self)
        self.input_thread_started = False

        # Command helpers
        self.skip_song = False  # Becomes True if the user enters the skip command
        self.exit_program = False  # Becomes True if the user enters the exit command
        self.loop_song = False  # Becomes True if the user enters the loop command
        self.shuffle_playlist = False  # Becomes True if the user enters the shuffle command
        self.back_song = False  # Becomes True if the user enters the back command
        self.back_amount = 1  # Stores how many indexes back will be popped from song_history to get songs in history


    def play_playlist_songs(self):
        """
        Play every song in the passed in playlist.
        Note: Skip is handled in command_handler and _get_next_song() is executed following a skip.
        """
        while True:
            if not self._input_features_enabled():  # No extra features enabled. Default.
                self._get_next_song()
            elif self.back_song:  # Back command entered
                self._get_next_song_back()
            elif self.loop_song:  # Looping enabled
                pass  # we don't need to change the value of self.current_song in this case
            elif self.shuffle_playlist:  # Shuffling enabled
                self.back_amount = 0
                self._get_next_song_shuffling()
            self._play_current_song()

    def _set_current_song(self, index):
        """
        Sets the current song to the index that was passed in.
        """
        self.current_song = self.playlist[index]

    def _get_next_song(self):
        """
        Sets the current song to the next song index and adds it to song_history.
        """
        self.back_amount = 0
        self._set_current_song(self.song_index)
        self._add_song_to_history()
        self.song_index += 1

    def _get_next_song_back(self):
        """
        Sets the current song to the last played song.
        Supports multiple back commands and shuffle.
        """
        self.back_amount -= 1
        if len(self.song_history) > abs(self.back_amount):
            self.song_index = self.song_history.pop(self.back_amount - 1)
            self._set_current_song(self.song_index)
            self._add_song_to_history()
        else:
            print("No songs remaining in history.")
        self.back_song = False

    def _get_next_song_shuffling(self):
        """
        Sets the current song to a random unique song in the playlist.
        Even YouTube couldn't write a better shuffling algorithm!
        """
        found_unique_random = False
        while not found_unique_random:
            random_int = random.randint(0, self.total_songs - 1)
            if random_int not in self.song_history:
                self.song_index = random_int
                self._set_current_song(random_int)
                self._add_song_to_history()
                found_unique_random = True
            elif self.song_counter == self.total_songs:
                self.song_history = [self.song_history[-1]]

    def _print_current_song_information(self, print_command_string=True):
        """
        Prints the current song's relevant information.
        """
        date = self.current_song.published.split()[0].split("-")
        year, month, day = date
        date = f"{month}/{day}/{year}"

        if self.song_info_enabled:
            os.system("cls||clear")
            print(             
f"""{Fore.CYAN}======================================
{Fore.GREEN}Title:{Fore.RESET} {self.current_song.title}
{Fore.GREEN}Length:{Fore.RESET} {self.current_song.duration}
{Fore.GREEN}Views:{Fore.RESET} {self.current_song.viewcount:,d}
{Fore.GREEN}Rating:{Fore.RESET} {round(self.current_song.rating, 2)}
{Fore.GREEN}Date:{Fore.RESET} {date}
{Fore.CYAN}======================================
{self.command_string if print_command_string else ""}""",
                end="",
            )

    def _play_current_song(self):
        """
        Plays the current song stored in self.current_song and adds it to song history.
        """

        self._reset_state()
        # Play song
        song_url = self.current_song.getbestaudio().url_https
        self.vlc_player.set_mrl(song_url, ":no-video")
        self.vlc_player.play()

        if not self.input_thread_started:
            self._print_current_song_information(print_command_string=False)
            print(f"{Fore.YELLOW}===Enter ? to view a list of commands==={Fore.RESET}")
            self.cmds.input_thread.start()
            self.input_thread_started = True
        else:
            self._print_current_song_information()

        # Sleep for duration of song
        self._song_timer()

    def _song_timer(self):
        """
        Loop and check for state change from user input while the song plays.
        """
        time.sleep(0.5)
        while self.vlc_player.is_playing() or self.vlc_is_paused():
            if self.skip_song:
                self.vlc_player.stop()
                self.skip_song = False
                break
            if self.exit_program:
                Fore.RESET
                Back.RESET
                sys.exit(0)
            time.sleep(0.5)

    def _reset_state(self):
        """
        Resets the song history to only contain the last song in the history if every song in the playlist has been played.
        """
        self.song_counter += 1
        if len(self.song_history) == self.total_songs:
            self.song_history = [self.song_history[-1]]
            self.song_counter = 0
            return True
        return False

    def _input_features_enabled(self):
        input_features = [self.loop_song, self.shuffle_playlist, self.back_song, self.skip_song]
        return True if True in input_features else False

    def _add_song_to_history(self):
        """
        Adds the current song index to song_history.
        """
        self.song_history.append(self.song_index)

    def vlc_is_paused(self):
        """
        Check if the VLC player is paused.
        """
        return True if self.vlc_player.get_state() == vlc.State.Paused else False


def is_valid_file(parser, arg):
    """
    Function used to check if a valid VLC path was given.
    """
    if not os.path.exists(arg):
        parser.error(
            f"The filepath {arg} does not exist! Be sure to include quotes around the path, view help for more info."
        )


def parse_args():
    """
    Parses passed in CLI arguments. 
    Output: tuple: YouTube playlist URL string, VLC Install Directory
    """

    parser = argparse.ArgumentParser(description="Streams YouTube Playlist in VLC")
    parser.add_argument(
        "youtube_playlist_URL",
        metavar="YouTube Playlist URL",
        help="URL to a YouTube Playlist",
    )
    parser.add_argument(
        "-v",
        metavar="VLC Install Directory",
        type=lambda x: is_valid_file(parser, x),
        help='If you\'re getting a FileNotFound error, use this option to pass in your VLC install directory. Example: "C:\\Program Files\\VideoLAN\VLC" Be sure to include the quotes.',
    )
    args = parser.parse_args()
    if args.v is None:
        args.v = "C:\Program Files\VideoLAN\VLC"
    return args.youtube_playlist_URL, args.v


def add_vlc_dir_to_path(vlc_dir):
    """
    Note: The python-vlc module depends on a .dll in the VLC install directory.
    Adds the install directory to the path and imports the vlc module.
    """
    app_dir = os.getcwd()
    os.add_dll_directory(vlc_dir)
    os.chdir(app_dir)
    global vlc
    import vlc


def read_playlist_and_VLC_url_from_file():
    """
    Reads previously stored YouTube playlist URL and VLC directory.
    Output: tuple: playlist url, vlc dir
    """
    url = ""
    vlc_dir = ""
    with open("data/playlist.txt", "r") as playlist_file:
        url = playlist_file.read()
    with open("data/vlc_dir.txt", "r") as vlc_dir_file:
        vlc_dir = vlc_dir_file.read()
    return url, vlc_dir


def write_playlist_url_and_vlc_dir_to_file(playlist_url, vlc_dir):
    """
    Stores the playlist url and vlc dir.
    """
    if not os.path.isdir("./data"):
        os.mkdir("./data")
    with open("./data/playlist.txt", "w+") as playlist_file:
        playlist_file.write(playlist_url)
    with open("./data/vlc_dir.txt", "w+") as vlc_dir_file:
        vlc_dir_file.write(vlc_dir)


def main():
    if len(sys.argv) == 1:
        if os.path.exists("data/playlist.txt") and os.path.exists("data/vlc_dir.txt"):
            youtube_playlist_URL, vlc_dir = read_playlist_and_VLC_url_from_file()
        else:
            print('No YouTube playlist stored. Run "python -m vlcyt -h" for help.')
            sys.exit(0)
    else:
        youtube_playlist_URL, vlc_dir = parse_args()
        write_playlist_url_and_vlc_dir_to_file(youtube_playlist_URL, vlc_dir)

    add_vlc_dir_to_path(vlc_dir)
    vlcyt = VLCYT(youtube_playlist_URL)
    vlcyt.play_playlist_songs()


if __name__ == "__main__":
    main()



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="VLCYT",
    version="1.6",  # update VLCYT version in requirements as well
    description="Stream your YouTube playlist in VLC behind the scenes from the command line.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Timothy Hill Jr",
    license="MIT",
    url="https://github.com/hillt03/VLCYT",
    keywords="youtube vlc stream playlist",
    packages=find_packages(),
    install_requires=[
        "colorama>=0.4.3",
        "pafy>=0.5.5",
        "python-vlc>=3.0.7110",
        "youtube-dl>=2019.11.28",
        "pyperclip>=1.7.0",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Microsoft :: Windows",
    ],
    python_requires=">=3.8",
)
